/*
 * SelectorThread.java
 *
 * Created on 24 de Maio de 2006, 11:41
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package br.ufsc.das.usage.nio;

import br.ufsc.das.usage.nio.CallbackErrorHandler;
import java.io.IOException;
import java.nio.channels.*;
import java.util.*;
import java.util.concurrent.Semaphore;

final public class SelectorThread implements Runnable {
    
    private Selector selector;
    private final Thread selectorThread;
    private Semaphore mutex = new Semaphore(1);
    
    private boolean closeRequested = false;
    private final List pendingInvocations = new ArrayList(32);
    
    
    public SelectorThread() throws IOException {
        selector = Selector.open();
        selectorThread = new Thread(this);
        selectorThread.start();
    }
    
    public void requestClose() {
        closeRequested = true;
        selector.wakeup();
    }
    
    
    public void addChannelInterestNow(SelectableChannel channel, int interest) throws IOException {
        if (Thread.currentThread() != selectorThread) {
            throw new IOException("Method can only be called from selector thread");
        }
        SelectionKey sk = channel.keyFor(selector);
        changeKeyInterest(sk, sk.interestOps() | interest);
    }
    
    public void addChannelInterestLater(final SelectableChannel channel,
            final int interest,
            final CallbackErrorHandler errorHandler) {
        
        invokeLater(new Runnable() {
            public void run() {
                try {
                    addChannelInterestNow(channel, interest);
                } catch (IOException e) {
                    if(errorHandler != null){
                        errorHandler.handleError(e);
                    }
                }
            }
        });
    }
    
    
    public void removeChannelInterestNow(SelectableChannel channel, int interest) throws IOException {
        if (Thread.currentThread() != selectorThread) {
            throw new IOException("Method can only be called from selector thread");
        }
        SelectionKey sk = channel.keyFor(selector);
        changeKeyInterest(sk, sk.interestOps() & ~interest);
    }
    
    public void removeChannelInterestLater(final SelectableChannel channel, final int interest,
            final CallbackErrorHandler errorHandler)  {
        invokeLater(new Runnable() {
            public void run() {
                try {
                    removeChannelInterestNow(channel, interest);
                } catch (IOException e) {
                    errorHandler.handleError(e);
                }
            }
        });
    }
    
    private void changeKeyInterest(SelectionKey sk, int newInterest) throws IOException {
        try {
            sk.interestOps(newInterest);
        } catch (CancelledKeyException cke) {
            IOException ioe = new IOException("Failed to change channel interest.");
            ioe.initCause(cke);
            throw ioe;
        }
    }
    
    public void registerChannelLater(final SelectableChannel channel, final int selectionKeys,
            final Object handlerInfo,
            final CallbackErrorHandler errorHandler) {
        invokeLater(new Runnable() {
            public void run() {
                try {
                    registerChannelNow(channel, selectionKeys, handlerInfo);
                } catch (IOException e) {
                    errorHandler.handleError(e);
                }
            }
        });
    }
    
    
    public void registerChannelNow(SelectableChannel channel, int selectionKeys,
            Object handlerInfo) throws IOException {
        if (Thread.currentThread() != selectorThread) {
            throw new IOException("Method can only be called from selector thread");
        }
        
        if (!channel.isOpen()) {
            throw new IOException("Channel is not open.");
        }
        
        try {
            if (channel.isRegistered()) {
                SelectionKey sk = channel.keyFor(selector);
                assert sk != null : "Channel is already registered with other selector";
                sk.interestOps(selectionKeys);
                Object previousAttach = sk.attach(handlerInfo);
                assert previousAttach != null;
            } else {
                channel.configureBlocking(false);
                channel.register(selector, selectionKeys, handlerInfo);
            }
        } catch (Exception e) {
            IOException ioe = new IOException("Error registering channel.");
            ioe.initCause(e);
            throw ioe;
        }
    }
    
    
    public void invokeLater(Runnable run) {
        try{
            this.mutex.acquire();
        }catch(Exception e){
            e.printStackTrace();
        }
        pendingInvocations.add(run);
        this.mutex.release();
        selector.wakeup();
    }
    
    public void invokeAndWait(final Runnable task) throws InterruptedException {
        if (Thread.currentThread() == selectorThread) {
            task.run();
        } else {
            final Semaphore sm = new Semaphore(0);
            this.invokeLater(new Runnable() {
                public void run() {
                    task.run();
                    sm.release();
                }
            });
            try{
                sm.acquire();
            }catch(Exception e){
                e.printStackTrace();
            }
        }
    }
    
    private void doInvocations() {
        try{
            this.mutex.acquire();
        }catch(Exception e){
            e.printStackTrace();
        }
        for (int i = 0; i < pendingInvocations.size(); i++) {
            Runnable task = (Runnable) pendingInvocations.get(i);
            task.run();
        }
        pendingInvocations.clear();
        this.mutex.release();
    }
    
    public void run() {
    
        while (true) {
            doInvocations();
            
            if (closeRequested) {
                return;
            }
            
            int selectedKeys = 0;
            try {
                selectedKeys = selector.select();
            } catch (IOException ioe) {
                ioe.printStackTrace();
                continue;
            }
            
            if (selectedKeys == 0) {
                 continue;
            }
           
            Iterator it = selector.selectedKeys().iterator();
            while (it.hasNext()) {
                SelectionKey sk = (SelectionKey)it.next();
                it.remove();
                try {
                    int readyOps = sk.readyOps();
                    sk.interestOps(sk.interestOps() & ~readyOps);

                    if (sk.isAcceptable()) {
                        ((Acceptor)sk.attachment()).handleAccept();
                    } else if (sk.isConnectable()) {
                        ((Connector)sk.attachment()).handleConnect();
                    } else {
                        PacketChannel rwHandler =
                                (PacketChannel)sk.attachment();
                        if (sk.isReadable()) {
                            rwHandler.handleRead();
                        }
                        if (sk.isValid() && sk.isWritable()) {
                             rwHandler.handleWrite();
                        }
                    }
                } catch (Throwable t) {
                    closeSelectorAndChannels();
                     return;
                }
            }
        }
    }
    
     private void closeSelectorAndChannels() {
        Set keys = selector.keys();
        for (Iterator iter = keys.iterator(); iter.hasNext();) {
            SelectionKey key = (SelectionKey)iter.next();
            try {
                key.channel().close();
            } catch (IOException ignore) {
 
            }
        }
        try {
            selector.close();
        } catch (IOException ignore) {
            
        }
    }
}